#!/usr/bin/env -S pytest -vvv


"""
Unit tests for rust binding test library

Run this test script to verify that functions can be compiled and run, and produce
expected results

n.b. The rust module needs to be compiled the first time this is run, but pytest will
hide the output of the rust compiler, so it may appear to hang for a little while.
Subsequent compilations should be much shorter
"""


import numpy as np
import numpy.lib.recfunctions as rf
import pytest

from oxvox.nns import OxVoxNNS


TEST_ARRAY = np.arange(9, dtype=np.float32).reshape((3, 3))
ORIGIN = np.array([0, 0, 0], dtype=np.float32)


TEST_SEARCH_POINTS = np.array(
    [
        [0.1, 0.1, 0.2],  # Point 0: Point 1's neighbour
        [0.2, 0.2, 0.1],  # Point 1: Point 0's neighbour
        [3.2, 1.2, 1.1],  # Point 2: A neighbour to Points 0 and 1 if r > 3 (ish)
        [3.3, 1.1, 1.0],  # Point 3: Similar to Point 2, also Point 2's neighbour
    ],
    dtype=np.float32,
)


def test_find_neighbours() -> None:
    """
    Test a simple case of finding neighbours in a small pointcloud
    """
    query_points = TEST_SEARCH_POINTS[0].reshape(1, -1)
    num_neighbours = 3
    max_dist = 4.0
    voxel_size = 0.3

    nns = OxVoxNNS(TEST_SEARCH_POINTS, max_dist)
    indices, distances = nns.find_neighbours(query_points, num_neighbours)

    assert np.all(indices == [0, 1, 2])
    assert np.allclose(distances, [0.0, 0.173, 3.410], atol=0.001)


def test_count_neighbours() -> None:
    """
    Test a simple case of counting neighbours in a small pointcloud
    """
    query_points = TEST_SEARCH_POINTS
    num_neighbours = 3
    max_dist = 2.0
    voxel_size = 0.3

    nns = OxVoxNNS(TEST_SEARCH_POINTS, 2.0)
    assert np.all(nns.count_neighbours(query_points) == [2] * 4)

    nns = OxVoxNNS(TEST_SEARCH_POINTS, 4.0)
    assert np.all(nns.count_neighbours(query_points) == [4] * 4)



#!/usr/bin/env python3

"""
Performance test the OxVox indexing module compared to native python and numpy_indexed
"""

import numpy as np
import numpy.lib.recfunctions as rfn
import numpy_indexed as npi
from time import time

from oxvox.indexing import indices_by_field


BIGASS_ARRAY = rfn.unstructured_to_structured(
    # np.random.randint(0, 100, (99999, 3)),
    np.random.randint(0, 100, (1_000_000, 3)),
    names=["a", "b", "c"],
)


def test_indices_by_field_naive():
    start = time()
    for unique_values in np.unique(BIGASS_ARRAY[["a", "b"]]):
        values = BIGASS_ARRAY[
            (BIGASS_ARRAY["a"] == unique_values[0])
            & (BIGASS_ARRAY["b"] == unique_values[1])
        ]
    print(f"Native indices_by_field took {time() - start}s")


def test_indices_by_field_npi():
    start = time()
    for chunk in npi.group_by(BIGASS_ARRAY[["a", "b"]]).split(BIGASS_ARRAY):
        pass
    print(f"Numpy-indexed indices_by_field took {time() - start}s")


def test_indices_by_field_oxvox():
    start = time()
    for unique_values, indices in indices_by_field(BIGASS_ARRAY, ["a", "b"]):
        pass
    print(f"OxVox indices_by_field took {time() - start}s")


if __name__ == "__main__":
    test_indices_by_field_naive()
    test_indices_by_field_npi()
    test_indices_by_field_oxvox()



#!/usr/bin/env python3


"""
Unit tests for rust binding test library

Run this test script to verify that functions can be compiled and run, and produce 
expected results

n.b. The rust module needs to be compiled the first time this is run, but pytest will
hide the output of the rust compiler, so it may appear to hang for a little while.
Subsequent compilations should be much shorter
"""


import numpy as np
from time import time
from typing import Tuple, Dict, Callable, Any
import numpy.typing as npt
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import json
import sys
import cloudpickle
import pickle
from colorsys import hsv_to_rgb
import tqdm
import numpy.lib.recfunctions as rf
from p_tqdm import p_imap
from functools import partial

# Competitor NNS algorithms
from scipy.spatial import KDTree as SciPyKDTree
from sklearn.neighbors import KDTree as SkLearnKDTree
import open3d as o3d

# Our NNS algorithm
from ox_vox_nns.ox_vox_nns import OxVoxNNS

from abyss.bedrock.io.convenience import easy_load

SEARCH_POINTS_IRL = rf.structured_to_unstructured(
    easy_load("/home/hmo/tmp/DEEPL-1947/4.bin")[["x", "y", "z"]]
).astype(np.float32)[::4]
SEARCH_POINTS_4M_UNIFORM = np.random.random((4_000_000, 3)).astype(np.float32) * 15
SEARCH_POINTS_4M_CLUSTERS = np.vstack(
    [
        # Generate a cluster of random points
        np.random.random((400_000, 3)).astype(np.float32) * 0.5
        # Move cluster somewhere randomly
        + np.random.random((1, 3)).astype(np.float32) * 15
    ]
    * 10
)
SEARCH_POINTS_1M_LOTS_CLUSTERS = np.vstack(
    [
        # Generate a cluster of random points
        np.random.random((10_000, 3)).astype(np.float32)
        # Move cluster somewhere randomly
        + np.random.random((1, 3)).astype(np.float32) * 15
    ]
    * 100
)


TEST_INPUTS = {
    "4m-uniform": {
        "search_points": SEARCH_POINTS_4M_UNIFORM,
        "query_points": SEARCH_POINTS_4M_UNIFORM,
    },
    # "4m-uniform-800": {
    #     "search_points": SEARCH_POINTS_4M_UNIFORM,
    #     "query_points": SEARCH_POINTS_4M_UNIFORM,
    #     "num_neighbours": 800,
    #     "max_dist": 0.05,
    #     "batch_size": 40_000,
    #     # "parallel": False,
    #     "parallel": True,
    #     "epsilon": 0.000,
    # },
    # "4m-clusters": {
    #     "search_points": SEARCH_POINTS_4M_CLUSTERS,
    #     "query_points": SEARCH_POINTS_4M_CLUSTERS,
    #     # "num_neighbours": 800,
    #     "num_neighbours": 8,
    #     "max_dist": 0.05,
    #     "voxel_size": 0.05,
    #     "batch_size": 40_000,
    # },
    # "4m-lots-clusters": {
    #     "search_points": SEARCH_POINTS_1M_LOTS_CLUSTERS,
    #     "query_points": SEARCH_POINTS_1M_LOTS_CLUSTERS,
    #     "num_neighbours": 800,
    #     # "num_neighbours": 8,
    #     "max_dist": 0.05,
    #     # "max_dist": 0.01,
    #     "parallel": True,
    #     "epsilon": 0.00,
    #     # "epsilon": 0.005,
    # },
    # "irl-1": {
    #     "search_points": SEARCH_POINTS_IRL,
    #     "query_points": SEARCH_POINTS_IRL,
    #     "num_neighbours": 1,
    #     "max_dist": 0.05,
    #     "voxel_size": 0.05,
    #     "batch_size": 40_000,
    #     "parallel": True,
    #     "epsilon": 0.000,
    # },
    # "irl-800": {
    #     "search_points": SEARCH_POINTS_IRL,
    #     "query_points": SEARCH_POINTS_IRL,
    #     "num_neighbours": 800,
    #     "max_dist": 0.05,
    #     "voxel_size": 0.05,
    #     "batch_size": 40_000,
    # },
}


def compare_performance_exact() -> int:
    """
    Compare performance of exact NNS algorithms
    """
    # Construct dict for output data
    results_exact = {}
    results_approx = {}

    for data_name, params in TEST_INPUTS.items():

        # Compare exact algos
        results_exact[data_name] = {}
        for algo_name, algo in {
            "scipy": _scipy_nns,
            "sklearn": _sklearn_nns,
            "Open3D": _o3d_nns,
            "OxVoxExact": partial(_oxvox_nns, sparse=False),
        }.items():
            results_exact[data_name][algo_name] = _run_test(
                data_name=data_name, algo_name=algo_name, algo=algo, params=params
            )
        with open(f"{data_name}_exact.html", mode="w", encoding="utf-8") as vis_fd:
            vis_fd.write(generate_plot(results_exact[data_name]))

        # Compare approximate algos
        results_approx[data_name] = {}
        for algo_name, algo in {
            "scipy_eps": partial(_scipy_nns, epsilon=0.005),
            "OxVoxApprox": partial(_oxvox_nns, sparse=False, epsilon=0.005),
            "OxVoxSparse": partial(_oxvox_nns, sparse=True),
        }.items():
            results_approx[data_name][algo_name] = _run_test(
                data_name=data_name, algo_name=algo_name, algo=algo, params=params
            )
        with open(f"{data_name}_approx.html", mode="w", encoding="utf-8") as vis_fd:
            vis_fd.write(generate_plot(results_approx[data_name]))

    # Dump results as json for later review
    print(json.dumps(results_exact) + "\n" * 3)
    print(json.dumps(results_approx))

    return 0


def _run_test(
    data_name: str,
    algo_name: str,
    algo: Callable,
    params: Dict[str, Any],
):
    """
    Run tests for a given algorithm and dataset, for various values of k and r
    """
    results = np.zeros((4, 4))
    for j, num_neighbours in enumerate(
        [
            1,
            #     10,
            #     100,
            #     300,
        ]
    ):
        for i, search_radius in enumerate(
            [
                0.001,
                # 0.005,
                # 0.1,
                # 0.2,
            ]
        ):
            print(
                f"Searching for {num_neighbours} nearest neighbours within a"
                f" {search_radius} unit radius in dataset {data_name} using"
                f" algorithm: {algo_name}... "
            )
            sys.stdout.flush()
            start = time()
            indices, distances = algo(
                **params, num_neighbours=num_neighbours, max_dist=search_radius
            )
            compute_time = time() - start
            print(f"Done in {compute_time}s")
            sys.stdout.flush()
            results[j, i] = compute_time
    return results


def generate_plot(test_results: Dict[str, npt.NDArray[np.floating]]) -> str:
    """
    Plot results of tests with plotly

    Args:
        test_results: Compute times from experiments, indexed by experiment name

    Returns:
        Plotly visualisation HTML contents
    """
    # fig = make_subplots(
    #     rows=1,
    #     cols=2,
    #     specs=[[{"is_3d": True}, {"is_3d": True}]],
    #     subplot_titles=[
    #         "Color corresponds to z",
    #         "Color corresponds to distance to origin",
    #     ],
    # )

    # fig.add_trace(go.Surface(x=x, y=y, z=z, colorbar_x=-0.07), 1, 1)
    # fig.add_trace(go.Surface(x=x, y=y, z=z, surfacecolor=x**2 + y**2 + z**2), 1, 2)
    # fig.update_layout(title_text="Ring cyclide")
    # fig.show()

    # Compute a distinct colour for each surface in the plot
    # colours = (
    #     np.array(hsv_to_rgb(h=hue, s=1, v=1)) * 255
    #     for hue in np.arange(0, 1, len(test_results))
    # )

    fig = go.Figure(
        data=[
            go.Surface(
                z=results,
                name=name,
                showscale=False,
                opacity=0.9,
                # surfacecolor=colour.tolist(),
            )
            for name, results in test_results.items()
        ]
    )
    fig.update_layout(
        scene={"xaxis_title": "Search radius", "yaxis_title": "Num neighbours"}
    )

    view = False
    embed_plotly = False
    output_path = None

    if view:
        fig.show()

    # Generate visualisation as html string
    html_str = fig.to_html(include_plotlyjs=True if embed_plotly else "cdn")

    # Make link to parent dir for navigation
    html_str = html_str.replace(
        "<body>",
        "<body><a href=.>Parent directory</a>\n",
    )

    # Dump to file
    if output_path:
        with open(output_path, mode="w", encoding="utf-8") as output_fd:
            output_fd.write(html_str)

    return html_str


"""
Wrappers around NNS implementations (with common usage semantics) below
"""


def _scipy_nns(
    search_points: npt.NDArray[np.float32],
    query_points: npt.NDArray[np.float32],
    num_neighbours: int,
    epsilon: float = 0,
    **kwargs,
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.float32]]:
    """
    Run nearest neighbour search using scipy
    """
    distances, indices = SciPyKDTree(search_points).query(
        query_points, k=num_neighbours, workers=-1, eps=epsilon
    )
    return indices, distances


def _sklearn_nns(
    search_points: npt.NDArray[np.float32],
    query_points: npt.NDArray[np.float32],
    num_neighbours: int,
    **kwargs,
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.float32]]:
    """
    Run nearest neighbour search using sklearn (single-threaded)
    """
    distances, indices = SkLearnKDTree(search_points, metric="euclidean").query(
        query_points, k=num_neighbours
    )
    return indices, distances


def _o3d_nns(
    search_points: npt.NDArray[np.float32],
    query_points: npt.NDArray[np.float32],
    num_neighbours: int,
    max_dist: float,
    **kwargs,
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.float32]]:
    """
    Run nearest neighbour search using open3d
    """
    nns = o3d.core.nns.NearestNeighborSearch(o3d.core.Tensor(search_points))
    nns.hybrid_index()

    # Construct generator of chunks of query points
    indices, distances, _ = nns.hybrid_search(
        query_points, radius=max_dist, max_knn=num_neighbours
    )
    return indices, distances


def _oxvox_nns(
    search_points: npt.NDArray[np.float32],
    query_points: npt.NDArray[np.float32],
    num_neighbours: int,
    max_dist: float,
    sparse: bool,
    num_threads: int = 0,
    epsilon: float = 0,
    l2_distance: bool = True,
    **kwargs,
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.float32]]:
    """
    Run nearest neighbour search using OxVoxNNS
    """
    start = time()
    nns = OxVoxNNS(search_points, max_dist)
    print(f"Constructed NNS searcher in {time() -start}s")
    return nns.find_neighbours(
        query_points, num_neighbours, sparse, l2_distance, num_threads, epsilon
    )


def _sklearn_nns_multiproc(
    search_points: npt.NDArray[np.float32],
    query_points: npt.NDArray[np.float32],
    num_neighbours: int,
    batch_size: int,
    **kwargs,
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.float32]]:
    """
    Run nearest neighbour search using sklearn (multiprocessed)
    """
    # Construct tree with search points
    tree = SkLearnKDTree(search_points, metric="euclidean")

    # Construct output arrays up front (we will fill them in in chunks)
    num_points = len(query_points)
    indices = np.full((num_points, num_neighbours), fill_value=-1)
    distances = np.full((num_points, num_neighbours), fill_value=-1)

    # Construct generator of chunks of query points
    query_chunk_offsets = range(0, len(query_points), batch_size)
    query_chunks = (query_points[i : i + batch_size, :] for i in query_chunk_offsets)

    # Map query across chunks of query points
    processed_chunks = p_imap(
        lambda query_chunk: tree.query(query_chunk, k=num_neighbours),
        query_chunks,
        tqdm=partial(tqdm.tqdm, disable=True),  # Disable tqdm bar
    )

    # Insert values back into output array
    for (chunk_indices, chunk_distances), chunk_offset in zip(
        processed_chunks, query_chunk_offsets
    ):
        indices[chunk_offset : chunk_offset + batch_size] = chunk_indices
        distances[chunk_offset : chunk_offset + batch_size] = chunk_distances

    return indices, distances


def _oxvox_nns_multiproc(
    search_points: npt.NDArray[np.float32],
    query_points: npt.NDArray[np.float32],
    num_neighbours: int,
    max_dist: float,
    batch_size: int,
    voxel_size: float,
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.float32]]:
    """
    Run nearest neighbour search using OxVox with Python multiprocessing
    """
    # Construct OxVoxNNS object with search points
    nns = OxVoxNNS(search_points, max_dist, voxel_size)

    # Construct output arrays up front (we will fill them in in chunks)
    num_points = len(query_points)
    indices = np.full((num_points, num_neighbours), fill_value=-1)
    distances = np.full((num_points, num_neighbours), fill_value=-1)

    # Construct generator of chunks of query points
    query_chunk_offsets = range(0, len(query_points), batch_size)
    query_chunks = (query_points[i : i + batch_size, :] for i in query_chunk_offsets)

    # Map query across chunks of query points
    processed_chunks = p_imap(
        lambda query_chunk: nns.find_neighbours(query_chunk, num_neighbours),
        query_chunks,
        tqdm=partial(tqdm.tqdm, disable=True),  # Disable tqdm bar
    )

    # Insert values back into output array
    for (chunk_indices, chunk_distances), chunk_offset in zip(
        processed_chunks, query_chunk_offsets
    ):
        indices[chunk_offset : chunk_offset + batch_size] = chunk_indices
        distances[chunk_offset : chunk_offset + batch_size] = chunk_distances

    return indices, distances


if __name__ == "__main__":
    sys.exit(compare_performance_exact())

    # sys.exit(
    #     sys.stdout.write(
    #         generate_plot(
    #             {
    #                 "foo": np.eye(4) + 3,
    #                 "bar": np.eye(4) + 4,
    #             }
    #         )
    #     )
    # )



#!/usr/bin/env -S pytest -vvv

"""
Unit tests for indexing operations
"""

import numpy as np
import numpy.lib.recfunctions as rf
import pytest

from oxvox.indexing import indices_by_field


TEST_ARRAY = np.array(
    [
        ("foo", 2, 0.5, 15),
        ("bar", 3, 1.5, 15),
        ("baz", 2, 2.5, 30),
        ("bat", 3, 3.5, 30),
        ("bet", 4, 4.5, 15),
        ("bot", 4, 5.5, 30),
    ],
    dtype=[
        ("field", "|O"),
        ("a", np.int32),
        ("b", np.float32),
        ("c", np.int32),
    ],
)


def test_indices_by_field() -> None:
    """
    Test that indices_by_field returns the correct row indices for each value in a given field
    """

    # Test a single field
    for computed, expected in zip(
        sorted(list(indices_by_field(arr=TEST_ARRAY, fields="a"))),
        [
            (2, [0, 2]),
            (3, [1, 3]),
            (4, [4, 5]),
        ],
    ):
        np.testing.assert_array_equal(computed[0], expected[0])
        np.testing.assert_array_equal(computed[1], expected[1])

    # Test multiple fields
    for computed, expected in zip(
        sorted(
            list(indices_by_field(arr=TEST_ARRAY, fields=["a", "c"])),
            key=lambda x: tuple(x[0]),
        ),
        [
            ((2, 15), [0]),
            ((2, 30), [2]),
            ((3, 15), [1]),
            ((3, 30), [3]),
            ((4, 15), [4]),
            ((4, 30), [5]),
        ],
    ):
        np.testing.assert_array_equal(tuple(computed[0]), expected[0])
        np.testing.assert_array_equal(tuple(computed[1]), expected[1])



"""
Python wrapper around Rust NNS engine for typing stubs and interpreter help
"""


from typing import Tuple
from functools import partial

import numpy as np
from numpy.lib.recfunctions import structured_to_unstructured
import numpy.typing as npt

from oxvox._oxvox import OxVoxNNSEngine


class OxVoxNNS:
    """
    A hybrid-ish nearest neighbour search implemented in rust, tailored towards
    consistent performance, especially on difficult inputs for KDTrees
    """

    def __init__(
        self, search_points: npt.NDArray[np.floating], search_radius: float
    ) -> None:
        """
        Construct neighbour searcher object

        Internally this groups search points by voxel and constructs a lookup for points
        by their voxel coordinates

        n.b. this class (and the rust object it constructs internally) can be pickled,
        allowing queries to be done in async/parallel contexts if required

        Args:
            search_points: Points to search for neighbours amongst. Can be given as a 2D
                unstructured (i.e. conventional) array with 3 columns, or as a
                structured array with "x", "y" and "z" columns at minimum
            search_radius: Maximum distance between points before they are no longer
                considered neighbours
        """
        # The rust engine strictly expects 3-column unstructured arrays of 32-bit
        # floats, so we must convert structured arrays to unstructured and enure we only
        # have 32-bit values
        search_points = self._sanitise_points(search_points)

        # Construct internal rust neighbour searcher
        self.engine = OxVoxNNSEngine(search_points, search_radius)

    def find_neighbours(
        self,
        query_points: npt.NDArray[np.floating],
        num_neighbours: int,
        num_threads: int = 0,
        # epsilon: float = np.finfo(np.float32).eps,
        epsilon: float = 0,
    ) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.float32]]:
        """
        Find neighbours in search points within range for all given query points

        Args:
            query_points: Points to search for neighbours of. Can be given as a 2D
                unstructured (i.e. conventional) array with 3 columns, or as a
                structured array with "x", "y" and "z" columns at minimum
            num_neighbours: Maximum number of neighbours to find, a.k.a. `k`
            num_threads: Number of parallel CPU threads to use in queries. Uses all
                available CPUs if set to 0
            epsilon: Any neighbours within this distance of the query point are accepted
                automatically (skips sorting). Even at its default value of float32 eps
                (0.00000012), this can help prevent the search from getting bogged down
                in extremely dense regions

        Returns:
            Indices of neighbouring search points. -1 where neighbours can't be found
            Distance to query point for each search point index. -1.0 where neighbours
                can't be found
        """
        return self.engine.find_neighbours(
            self._sanitise_points(query_points),
            num_neighbours,
            num_threads,
            epsilon,
        )

    def count_neighbours(
        self,
        query_points: npt.NDArray[np.floating],
        num_threads: int = 0,
    ) -> npt.NDArray[np.uint32]:
        """
        Count neighbours in search points within range for all given query points

        Args:
            query_points: Points to search for neighbours of. Can be given as a 2D
                unstructured (i.e. conventional) array with 3 columns, or as a
                structured array with "x", "y" and "z" columns at minimum
            num_neighbours: Maximum number of neighbours to find, a.k.a. `k`
            num_threads: Number of parallel CPU threads to use in queries. Uses all
                available CPUs if set to 0
            epsilon: Any neighbours within this distance of the query point are accepted
                automatically (skips sorting). Even at its default value of float32 eps
                (0.00000012), this can help prevent the search from getting bogged down
                in extremely dense regions

        Returns:
            Indices of neighbouring search points. -1 where neighbours can't be found
            Distance to query point for each search point index. -1.0 where neighbours
                can't be found
        """
        return self.engine.count_neighbours(
            self._sanitise_points(query_points),
            num_threads,
        )

    @staticmethod
    def _sanitise_points(points: npt.NDArray[np.floating]) -> npt.NDArray[np.float32]:
        """
        Prepare pointcloud arrays to be used by rust engine, which expects arrays of
        32-bit floats

        Args:
            points: Pointcloud to be sanitised

        Returns:
            Pointcloud, ready to be passed into rust engine
        """
        return (
            structured_to_unstructured(points[["x", "y", "z"]], dtype=np.float32)
            if points.dtype.names is not None
            else points.astype(np.float32)
        )



"""
Array indexing operations
"""

from typing import Any, Iterator

import numpy as np
import numpy.typing as npt
import numpy_indexed as npi

from oxvox._oxvox import indices_by_field as indices_by_field_rust


def indices_by_field(
    arr: npt.NDArray[Any], fields: str | tuple[str, ...]
) -> Iterator[tuple[Any, npt.NDArray[np.uint64]]]:
    """
    Compute row indices for each value in a given field in a structured array

    This can be useful for modifying an array in-place. If this is not a requirement,
    consider using the numpy_indexed library directly (used internally by this function)

    Args:
        arr: Structured array to compute row indices for
        fields: Field(s) to split by

    Yields:
        Value(s) in the given field(s) in the input array
        Row indices with those values

    Examples:

        The following snippet modifies in-place, but requires a full pass through the
        array for each unique value, making it O(n*u) where n is the number of rows and
        u is the number of unique values:

            for value in np.unique(arr[fields]):
                arr[arr[fields] == value] = some_updated_value

        Similarly, we can use numpy_indexed to do this in O(n+u), so long as we accept
        the output array being sorted by the fields we are splitting by:

            chunks = []
            for chunk in npi.group_by(arr[fields]).split(arr):
                chunk["some_field"] = some_updated_value
                chunks.append(chunk)
            return np.concatenate(chunks)

        With oxvox we can do this in O(n+u) time, in parallel, and without sorting the
        output array:

            for value, indices in indices_by_field(arr, fields):
                arr[indices] = some_updated_value

        n.b. If you don't need the values in the array, and are happy with unique
        sequential IDs as dictionary keys, you can use the underlying
        `indices_by_field_rust` directly:

            for offset, indices in enumerate(
                indices_by_field_rust(unique_ids, counts).values()
            ):
                arr[indices] += offset
    """

    # We first use numpy_indexed (because it handles structured arrays nicely) to get:
    # - An array the same length as the input array, with a unique ID for each unique
    #       set of values in the given field(s)
    # - A set of unique values in the given field(s)
    # - A count of how many times each unique value appears in the given field(s)
    grouper = npi.group_by(arr[fields])
    unique_ids = grouper.inverse
    unique_values = grouper.unique
    counts = grouper.count

    # We quickly create a lookup from unique IDs to their unique array values
    index_to_value = dict(enumerate(unique_values))

    # Now we use the rust engine to compute the row indices for each unique ID
    for unique_id, indices in indices_by_field_rust(unique_ids, counts).items():
        yield index_to_value[unique_id], indices
